/******************************************************************************
 *                           O P E N / C A E S A R
 *-----------------------------------------------------------------------------
 *   INRIA - Unite de Recherche Rhone-Alpes
 *   655, avenue de l'Europe
 *   38330 Montbonnot Saint Martin
 *   FRANCE
 *-----------------------------------------------------------------------------
 *   Module             :       generator.c
 *   Auteur             :       Hubert GARAVEL
 *   Version            :       3.8
 *   Date               :       06/07/25 12:12:09
 *****************************************************************************/

char caesar_sccs_what_generator[] =
"@(#)open/caesar -- 3.8 -- 06/07/25 12:12:09 -- generator.c";

char *LDFLAGS = "@(#)LDFLAGS = -L\"${BCG:-$CADP}\"/bin.`\"$CADP\"/com/arch` -lBCG_IO -lBCG -lm";

#include "caesar_graph.h"
#include "caesar_edge.h"
#include "caesar_table_1.h"
#include "caesar_mask_1.h"
#include "caesar_hide_1.h"

#include "bcg_user.h"

/*---------------------------------------------------------------------------*/

static void caesar_abort (caesar_signal)
int  caesar_signal;

{
     caesar_use (caesar_signal);
     CAESAR_SIGNALS (SIG_IGN);
     BCG_IO_WRITE_BCG_END ();
     CAESAR_ERROR ("caught interrupt: closing BCG file");
}

/*---------------------------------------------------------------------------*/

int  main (argc, argv)
int  argc;
char *argv[];

{
     char *caesar_filename;
     CAESAR_TYPE_INTEGER caesar_nb_tokens;
     CAESAR_TYPE_MASK_1 caesar_m;
     CAESAR_TYPE_BOOLEAN caesar_monitor;
     CAESAR_TYPE_TABLE_1 caesar_t;
     CAESAR_TYPE_STRING caesar_comment;
     CAESAR_TYPE_STATE caesar_s1, caesar_s2;
     CAESAR_TYPE_EDGE caesar_e1_en, caesar_e;
     CAESAR_TYPE_LABEL caesar_l;
     CAESAR_TYPE_STRING caesar_label;
     CAESAR_TYPE_INDEX_TABLE_1 caesar_n1, caesar_n2;
     CAESAR_TYPE_POINTER caesar_dummy;
     CAESAR_TYPE_INDEX_TABLE_1 caesar_initial_state;
     /* PC */
     CAESAR_TYPE_BOOLEAN caesar_rate;
     float caesar_curr_rate;
     float caesar_total_rate;
     float caesar_unif_rate;
     CAESAR_TYPE_BOOLEAN caesar_stable_state;
     CAESAR_TYPE_BOOLEAN caesar_io;
     CAESAR_TYPE_HIDE_1 caesar_output;
     CAESAR_TYPE_STRING caesar_new_label;
     /* PC */

     /* initializing BCG and OPEN/CAESAR */

     BCG_INIT ();
     BCG_INIT_OPTIONS ();
     CAESAR_INIT_GRAPH ();

     /* parsing command-line options */

     CAESAR_TOOL = argv[0];
     --argc;
     ++argv;
     caesar_monitor = CAESAR_FALSE;
     /* PC */
     caesar_rate = CAESAR_FALSE;
     caesar_io = CAESAR_FALSE;
     /* PC */
     CAESAR_CREATE_MASK_1 (&caesar_m, CAESAR_LABEL_AREA_1 (), CAESAR_TRUE, 0, CAESAR_TRUE, NULL, NULL, NULL, CAESAR_TRUE);
     while (argc > 0) {
	  /* parsing general BCG options (see "man bcg") */
	  caesar_nb_tokens = BCG_PARSE_OPTION (argc, argv, BCG_WRITE_OPTIONS);
	  if (caesar_nb_tokens != 0) {
	       /* a general BCG option was recognized */
	       argc -= caesar_nb_tokens;
	       argv += caesar_nb_tokens;
	       continue;
	  }
	  /* parsing hiding and renaming options (see "man caesar_mask_1") */
	  CAESAR_PARSE_OPTION_MASK_1 (caesar_m, argc, argv, &caesar_nb_tokens);
	  if (caesar_nb_tokens < 0)
	       CAESAR_ERROR ("invalid hiding or renaming option");
	  else if (caesar_nb_tokens > 0) {
	       /* one or more hiding or renaming options were recognized */
	       argc -= caesar_nb_tokens;
	       argv += caesar_nb_tokens;
	       continue;
	  }
	  if (strcmp (*argv, "-monitor") == 0) {
	       caesar_monitor = CAESAR_TRUE;
	       --argc;
	       ++argv;
	       continue;
	  /* PC */
	  } else if (strcmp (*argv, "-rate") == 0) {
	       caesar_rate = CAESAR_TRUE;
	       --argc;
	       ++argv;
          if ((argc < 1) || (sscanf(*argv,"%f",&caesar_unif_rate)<1)) {
               CAESAR_ERROR("rate option requires a uniformization rate");
          }
          caesar_new_label = (char*) calloc(30,sizeof(char));
	       --argc;
	       ++argv;
	       continue;
	  } else if (strcmp (*argv, "-io") == 0) {
	       caesar_io = CAESAR_TRUE;
               --argc;
               ++argv;
               if (argc < 1) {
                  CAESAR_ERROR("output option requires a filename");
               }
	       CAESAR_CREATE_HIDE_1(&caesar_output, argv[0], "output", "input", 0);
               if(caesar_output == NULL)
	          CAESAR_ERROR ("not enough memory for input/output mask");
               --argc;
               ++argv;
               continue;
	  /* PC */
	  } else if (*argv[0] == '-') {
	       CAESAR_ERROR ("unknown option ``%s''", *argv);
	  } else {
	       /* found a string corresponding to a file name */
	       break;
	  }
     }
     if (argc != 1)
	  CAESAR_ERROR ("one BCG filename expected as argument");
     caesar_filename = argv[0];

     /* allocating edges and state table */

     CAESAR_INIT_EDGE (CAESAR_FALSE, CAESAR_TRUE, CAESAR_TRUE, 0, 0);
     CAESAR_CREATE_TABLE_1 (&caesar_t, CAESAR_STATE_AREA_1 (), CAESAR_EMPTY_AREA_1 (), 0, 0, CAESAR_TRUE, NULL, NULL, NULL, NULL);
     if (caesar_t == NULL)
	  CAESAR_ERROR ("not enough memory for table");

     /* computing and storing the initial state */

     CAESAR_START_STATE ((CAESAR_TYPE_STATE) CAESAR_PUT_BASE_TABLE_1 (caesar_t));
     CAESAR_PUT_TABLE_1 (caesar_t);
     caesar_initial_state = CAESAR_GET_INDEX_TABLE_1 (caesar_t);

     /* opening the BCG graph for writing */

     caesar_comment = CAESAR_HISTORY_MASK_1 (caesar_m, "created by generator", NULL, NULL, NULL, NULL);
     BCG_IO_WRITE_BCG_BEGIN (caesar_filename, caesar_initial_state, 2, caesar_comment, caesar_monitor);
     CAESAR_DELETE (caesar_comment);

     CAESAR_SIGNALS (caesar_abort);

     /* main program loop */
     while (!CAESAR_EXPLORED_TABLE_1 (caesar_t)) {
	  caesar_s1 = (CAESAR_TYPE_STATE) CAESAR_GET_BASE_TABLE_1 (caesar_t);
	  caesar_n1 = CAESAR_GET_INDEX_TABLE_1 (caesar_t);
	  CAESAR_GET_TABLE_1 (caesar_t);

	  CAESAR_CREATE_EDGE_LIST (caesar_s1, &caesar_e1_en, 1);
	  if (CAESAR_TRUNCATION_EDGE_LIST () != 0)
	       CAESAR_ERROR ("not enough memory for edge lists");

          /* PC */
          if(caesar_rate){
		  caesar_stable_state = CAESAR_TRUE;
		  CAESAR_ITERATE_LN_EDGE_LIST (caesar_e1_en, caesar_e, caesar_l, caesar_s2) {
		       caesar_label = CAESAR_APPLY_MASK_1 (caesar_m, (CAESAR_TYPE_POINTER) caesar_l);
		       caesar_stable_state = CAESAR_VISIBLE_LABEL (caesar_l) && ((!caesar_io) || (!CAESAR_TEST_HIDE_1(caesar_output,CAESAR_STRING_LABEL(caesar_l))));
	               if (!caesar_stable_state)
	        	  break; 
	  	  }
          }
	  /* PC */

	  /* PC */
	  caesar_total_rate=0;
	  CAESAR_ITERATE_LN_EDGE_LIST (caesar_e1_en, caesar_e, caesar_l, caesar_s2) {
	       caesar_label = CAESAR_APPLY_MASK_1 (caesar_m, (CAESAR_TYPE_POINTER) caesar_l);
               /* PC */
               if((!caesar_rate) || caesar_stable_state || (strncmp(CAESAR_STRING_LABEL(caesar_l),"rate",4)!=0)){
               /* PC */
	               CAESAR_COPY_STATE ((CAESAR_TYPE_STATE) 
                  CAESAR_PUT_BASE_TABLE_1 (caesar_t), caesar_s2);
	               (void) CAESAR_SEARCH_AND_PUT_TABLE_1 (caesar_t, &caesar_n2, &caesar_dummy);
	               BCG_IO_WRITE_BCG_EDGE (caesar_n1, caesar_label, caesar_n2);
                  if((strncmp(CAESAR_STRING_LABEL(caesar_l),"rate",4)==0) && (sscanf(CAESAR_STRING_LABEL(caesar_l),"%*s %f",&caesar_curr_rate)==1)){
                    caesar_total_rate=caesar_total_rate+caesar_curr_rate;
                  }
               /* PC */
               }
               /* PC */
	  }
	  CAESAR_DELETE_EDGE_LIST (&caesar_e1_en);
     /* PC */
     if(caesar_rate && (caesar_stable_state) && (caesar_total_rate<caesar_unif_rate)){
       sprintf(caesar_new_label,"rate %f",caesar_unif_rate-caesar_total_rate);
       BCG_IO_WRITE_BCG_EDGE (caesar_n1, caesar_new_label, caesar_n1);
     }
     /* PC */
     }

     /* closing the BCG file and terminating */

     CAESAR_DELETE_TABLE_1 (&caesar_t);
     CAESAR_DELETE_MASK_1 (&caesar_m);
     /* PC */
     if(caesar_io)
          CAESAR_DELETE_HIDE_1 (&caesar_output);
     if(caesar_rate)
          free(caesar_new_label);
     /* PC */
     BCG_IO_WRITE_BCG_END ();

     return (0);
}

/*---------------------------------------------------------------------------*/
